"""
Utility functions for binning
"""

import numpy as np
from scipy.interpolate import interp1d
import astropy.units as u
from astropy.table import QTable


#: Index returned by `calculate_bin_indices` for underflown values
UNDERFLOW_INDEX = np.iinfo(np.int64).min
#: Index returned by `calculate_bin_indices` for overflown values
OVERFLOW_INDEX = np.iinfo(np.int64).max


def bin_center(edges):
    return 0.5 * (edges[:-1] + edges[1:])


def join_bin_lo_hi(bin_lo, bin_hi):
    """
    Function joins bins into lo and hi part,
    e.g. [0, 1, 2] and [1, 2, 4] into [0, 1, 2, 4]
    It works on multidimentional arrays as long as the binning is in the last axis

    Parameters
    ----------
    bin_lo: np.array or u.Quantity
        Lo bin edges array
    bin_hi: np.array or u.Quantity
        Hi bin edges array

    Returns
    -------
    bins: np.array of u.Quantity
        The joint bins
    """

    if np.allclose(bin_lo[...,1:], bin_hi[...,:-1], rtol=1.e-5):
        last_axis=len(bin_lo.shape)-1
        bins = np.concatenate((bin_lo, bin_hi[...,-1:]), axis=last_axis)
        return bins
    else:
        raise ValueError('Not matching bin edges')


def split_bin_lo_hi(bins):
    """
    Inverted function to join_bin_hi_lo,
    e.g. it splits [0, 1, 2, 4] into [0, 1, 2] and [1, 2, 4]

    Parameters
    ----------
    bins: np.array of u.Quantity
        The joint bins

    Returns
    -------
    bin_lo: np.array or u.Quantity
        Lo bin edges array
    bin_hi: np.array or u.Quantity
        Hi bin edges array
    """
    bin_lo=bins[...,:-1]
    bin_hi=bins[...,1:]
    return bin_lo, bin_hi

def add_overflow_bins(bins, positive=True):
    """
    Add under and overflow bins to a bin array.

    Parameters
    ----------
    bins: np.array or u.Quantity
        Bin edges array
    positive: bool
        If True, the underflow array will start at 0, if not at ``-np.inf``
    """
    lower = 0 if positive else -np.inf
    upper = np.inf

    if hasattr(bins, "unit"):
        lower *= bins.unit
        upper *= bins.unit

    if bins[0] > lower:
        bins = np.append(lower, bins)

    if bins[-1] < upper:
        bins = np.append(bins, upper)

    return bins


@u.quantity_input(e_min=u.TeV, e_max=u.TeV)
def create_bins_per_decade(e_min, e_max, bins_per_decade=5):
    """
    Create a bin array with bins equally spaced in logarithmic energy
    with ``bins_per_decade`` bins per decade.

    Parameters
    ----------
    e_min: u.Quantity[energy]
        Minimum energy, inclusive
    e_max: u.Quantity[energy]
        Maximum energy, non-inclusive
        If the endpoint exactly matches the ``n_bins_per_decade`` requirement,
        it will be included.
    n_bins_per_decade: int
        number of bins per decade

    Returns
    -------
    bins: u.Quantity[energy]
        The created bin array, will have units of e_min

    """
    unit = e_min.unit
    log_lower = np.log10(e_min.to_value(unit))
    log_upper = np.log10(e_max.to_value(unit))

    step = 1 / bins_per_decade
    # include endpoint if reasonably close
    eps = step / 10000
    bins = 10 ** np.arange(log_lower, log_upper + eps, step)
    return u.Quantity(bins, e_min.unit, copy=False)


def calculate_bin_indices(data, bins):
    """
    Calculate bin indices for given data array and bins.
    Underflow will be `UNDERFLOW_INDEX` and overflow `OVERFLOW_INDEX`.

    If the bins already include underflow / overflow bins, e.g.
    `bins[0] = -np.inf` and `bins[-1] = np.inf`, using the result of this
    function will always be a valid index into the resulting histogram.


    Parameters
    ----------
    data: ``~np.ndarray`` or ``~astropy.units.Quantity``
        Array with the data

    bins: ``~np.ndarray`` or ``~astropy.units.Quantity``
        Array or Quantity of bin edges. Must have the same unit as ``data`` if a Quantity.


    Returns
    -------
    bin_index: np.ndarray[int]
        Indices of the histogram bin the values in data belong to.
        Under- and overflown values will have values of `UNDERFLOW_INDEX`
        and `OVERFLOW_INDEX` respectively.
    """

    if hasattr(data, "unit"):
        if not hasattr(bins, "unit"):
            raise TypeError(f"If ``data`` is a Quantity, so must ``bins``, got {bins}")
        unit = data.unit
        data = data.to_value(unit)
        bins = bins.to_value(unit)

    n_bins = len(bins) - 1
    idx = np.digitize(data, bins) - 1

    underflow = (idx < 0)
    overflow = (idx >= n_bins)
    idx[underflow] = UNDERFLOW_INDEX
    idx[overflow] = OVERFLOW_INDEX
    valid = ~underflow & ~overflow
    return idx, valid


def create_histogram_table(events, bins, key="reco_energy"):
    """
    Histogram a variable from events data into an astropy table.

    Parameters
    ----------
    events : ``astropy.QTable``
        Astropy Table object containing the reconstructed events information.
    bins: ``~np.ndarray`` or ``~astropy.units.Quantity``
        Array or Quantity of bin edges.
        It must have the same units as ``data`` if a Quantity.
    key : ``string``
        Variable to histogram from the events table.

    Returns
    -------
    hist: ``astropy.QTable``
        Astropy table containg the histogram.
    """
    hist = QTable()
    hist[key + "_low"] = bins[:-1]
    hist[key + "_high"] = bins[1:]
    hist[key + "_center"] = 0.5 * (hist[key + "_low"] + hist[key + "_high"])
    hist["n"], _ = np.histogram(events[key], bins)

    # also calculate weighted number of events
    if "weight" in events.colnames:
        hist["n_weighted"], _ = np.histogram(
            events[key], bins, weights=events["weight"]
        )
    else:
        hist["n_weighted"] = hist["n"]

    # create counts per particle type, only works if there is at least 1 event
    if 'particle_type' in events.colnames and len(events) > 0:
        by_particle = events.group_by('particle_type')

        for group_key, group in zip(by_particle.groups.keys, by_particle.groups):
            particle = group_key['particle_type']

            hist["n_" + particle], _ = np.histogram(group[key], bins)

            # also calculate weighted number of events
            col = "n_" + particle
            if "weight" in events.colnames:
                hist[col + "_weighted"], _ = np.histogram(
                    group[key], bins, weights=group["weight"]
                )
            else:
                hist[col + "_weighted"] = hist[col]

    return hist


def resample_histogram1d(data, old_edges, new_edges, axis=0):
    """
    Rebinning of a histogram by interpolation along a given axis.

    Parameters
    ----------
    data : ``numpy.ndarray`` or ``astropy.units.Quantity``
        Histogram.
    old_edges : ``numpy.array`` or ``astropy.units.Quantity``
        Binning used to calculate ``data``.
        ``len(old_edges) - 1`` needs to equal the length of ``data``
        along interpolation axis (``axis``).
        If quantity, needs to be compatible to ``new_edges``.
    new_edges : ``numpy.array`` or ``astropy.units.Quantity``
        Binning of new histogram.
        If quantity, needs to be compatible to ``old_edges``.
    axis : int
        Interpolation axis.

    Returns
    -------
    ``numpy.ndarray`` or ``astropy.units.Quantity``
        Interpolated histogram with dimension according to ``data`` and ``new_edges``.
        If ``data`` is a quantity, this has the same unit.
    """

    data_unit = None
    if isinstance(data, u.Quantity):
        data_unit = data.unit
        data = data.to_value(data_unit)

    over_underflow_bin_width = old_edges[-2] - old_edges[1]
    old_edges = u.Quantity(
        np.nan_to_num(
            old_edges,
            posinf=old_edges[-2] + over_underflow_bin_width,
            neginf=old_edges[1] - over_underflow_bin_width,
        )
    )

    new_edges = u.Quantity(np.nan_to_num(new_edges))

    old_edges = old_edges.to(new_edges.unit)

    cumsum = np.insert(data.cumsum(axis=axis), 0, 0, axis=axis)

    norm = data.sum(axis=axis, keepdims=True)
    norm[norm == 0] = 1
    cumsum /= norm

    f_integral = interp1d(
        old_edges, cumsum, bounds_error=False,
        fill_value=(0, 1), kind="quadratic",
        axis=axis,
    )

    values = np.diff(f_integral(new_edges), axis=axis) * norm
    if data_unit:
        values = u.Quantity(values, unit=data_unit)

    return values



import astropy.units as u
import numpy as np
from astropy.table import QTable


def test_effective_area():
    from pyirf.irf import effective_area

    n_selected = np.array([10, 20, 30])
    n_simulated = np.array([100, 2000, 15000])

    area = 1e5 * u.m ** 2

    assert u.allclose(
        effective_area(n_selected, n_simulated, area), [1e4, 1e3, 200] * u.m ** 2
    )


def test_effective_area_per_energy():
    from pyirf.irf import effective_area_per_energy
    from pyirf.simulations import SimulatedEventsInfo

    true_energy_bins = [0.1, 1.0, 10.0] * u.TeV
    selected_events = QTable(
        {
            "true_energy": np.append(np.full(1000, 0.5), np.full(10, 5)),
        }
    )

    # this should give 100000 events in the first bin and 10000 in the second
    simulation_info = SimulatedEventsInfo(
        n_showers=110000,
        energy_min=true_energy_bins[0],
        energy_max=true_energy_bins[-1],
        max_impact=100 / np.sqrt(np.pi) * u.m,  # this should give a nice round area
        spectral_index=-2,
        viewcone=0 * u.deg,
    )

    area = effective_area_per_energy(selected_events, simulation_info, true_energy_bins)

    assert area.shape == (len(true_energy_bins) - 1,)
    assert area.unit == u.m ** 2
    assert u.allclose(area, [100, 10] * u.m ** 2)


def test_effective_area_energy_fov():
    from pyirf.irf import effective_area_per_energy_and_fov
    from pyirf.simulations import SimulatedEventsInfo

    true_energy_bins = [0.1, 1.0, 10.0] * u.TeV
    # choose edges so that half are in each bin in fov
    fov_offset_bins = [0, np.arccos(0.98), np.arccos(0.96)] * u.rad
    center_1, center_2 = 0.5 * (fov_offset_bins[:-1] + fov_offset_bins[1:]).to_value(
        u.deg
    )

    selected_events = QTable(
        {
            "true_energy": np.concatenate(
                [
                    np.full(1000, 0.5),
                    np.full(10, 5),
                    np.full(500, 0.5),
                    np.full(5, 5),
                ]
            )
            * u.TeV,
            "true_source_fov_offset": np.append(
                np.full(1010, center_1), np.full(505, center_2)
            )
            * u.deg,
        }
    )

    # this should give 100000 events in the first bin and 10000 in the second
    simulation_info = SimulatedEventsInfo(
        n_showers=110000,
        energy_min=true_energy_bins[0],
        energy_max=true_energy_bins[-1],
        max_impact=100 / np.sqrt(np.pi) * u.m,  # this should give a nice round area
        spectral_index=-2,
        viewcone=fov_offset_bins[-1],
    )

    area = effective_area_per_energy_and_fov(
        selected_events, simulation_info, true_energy_bins, fov_offset_bins
    )

    assert area.shape == (len(true_energy_bins) - 1, len(fov_offset_bins) - 1)
    assert area.unit == u.m ** 2
    assert u.allclose(area[:, 0], [200, 20] * u.m ** 2)
    assert u.allclose(area[:, 1], [100, 10] * u.m ** 2)



"""
Functions and classes for calculating spectral weights
"""
import astropy.units as u
import numpy as np
from scipy.interpolate import interp1d
from pkg_resources import resource_filename
from astropy.table import QTable

from .utils import cone_solid_angle

#: Unit of a point source flux
#:
#: Number of particles per Energy, time and area
POINT_SOURCE_FLUX_UNIT = (1 / u.TeV / u.s / u.m ** 2).unit

#: Unit of a diffuse flux
#:
#: Number of particles per Energy, time, area and solid_angle
DIFFUSE_FLUX_UNIT = POINT_SOURCE_FLUX_UNIT / u.sr


__all__ = [
    "POINT_SOURCE_FLUX_UNIT",
    "DIFFUSE_FLUX_UNIT",
    "calculate_event_weights",
    "PowerLaw",
    "LogParabola",
    "PowerLawWithExponentialGaussian",
    "CRAB_HEGRA",
    "CRAB_MAGIC_JHEAP2015",
    "PDG_ALL_PARTICLE",
    "IRFDOC_PROTON_SPECTRUM",
    "IRFDOC_ELECTRON_SPECTRUM",
    "TableInterpolationSpectrum",
    "DAMPE_P_He_SPECTRUM",
]


@u.quantity_input(true_energy=u.TeV)
def calculate_event_weights(true_energy, target_spectrum, simulated_spectrum):
    r"""
    Calculate event weights

    Events with a certain ``simulated_spectrum`` are reweighted to ``target_spectrum``.

    .. math::
        w_i = \frac{\Phi_\text{Target}(E_i)}{\Phi_\text{Simulation}(E_i)}

    Parameters
    ----------
    true_energy: astropy.units.Quantity[energy]
        True energy of the event
    target_spectrum: callable
        The target spectrum. Must be a allable with signature (energy) -> flux
    simulated_spectrum: callable
        The simulated spectrum. Must be a callable with signature (energy) -> flux

    Returns
    -------
    weights: numpy.ndarray
        Weights for each event
    """
    return (target_spectrum(true_energy) / simulated_spectrum(true_energy)).to_value(
        u.one
    )


class PowerLaw:
    r"""
    A power law with normalization, reference energy and index.
    Index includes the sign:

    .. math::

        \Phi(E, \Phi_0, \gamma, E_\text{ref}) =
        \Phi_0 \left(\frac{E}{E_\text{ref}}\right)^{\gamma}

    Attributes
    ----------
    normalization: astropy.units.Quantity[flux]
        :math:`\Phi_0`,
    index: float
        :math:`\gamma`
    e_ref: astropy.units.Quantity[energy]
        :math:`E_\text{ref}`
    """

    @u.quantity_input(
        normalization=[DIFFUSE_FLUX_UNIT, POINT_SOURCE_FLUX_UNIT], e_ref=u.TeV
    )
    def __init__(self, normalization, index, e_ref=1 * u.TeV):
        """Create a new PowerLaw spectrum"""
        if index > 0:
            raise ValueError(f'Index must be < 0, got {index}')

        self.normalization = normalization
        self.index = index
        self.e_ref = e_ref

    @u.quantity_input(energy=u.TeV)
    def __call__(self, energy):
        e = (energy / self.e_ref).to_value(u.one)
        return self.normalization * e**self.index

    @classmethod
    @u.quantity_input(obstime=u.hour, e_ref=u.TeV)
    def from_simulation(cls, simulated_event_info, obstime, e_ref=1 * u.TeV):
        """
        Calculate the flux normalization for simulated events drawn
        from a power law for a certain observation time.
        """
        e_min = simulated_event_info.energy_min
        e_max = simulated_event_info.energy_max
        index = simulated_event_info.spectral_index
        n_showers = simulated_event_info.n_showers
        viewcone = simulated_event_info.viewcone

        if viewcone.value > 0:
            solid_angle = 2 * np.pi * (1 - np.cos(viewcone)) * u.sr
            unit = DIFFUSE_FLUX_UNIT
        else:
            solid_angle = 1
            unit = POINT_SOURCE_FLUX_UNIT

        A = np.pi * simulated_event_info.max_impact ** 2

        delta = e_max ** (index + 1) - e_min ** (index + 1)
        nom = (index + 1) * e_ref ** index * n_showers
        denom = (A * obstime * solid_angle) * delta

        return cls(normalization=(nom / denom).to(unit), index=index, e_ref=e_ref,)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.normalization} * (E / {self.e_ref})**{self.index})"

    @u.quantity_input(inner=u.rad, outer=u.rad)
    def integrate_cone(self, inner, outer):
        """Integrate this powerlaw over solid angle in the given cone

        Parameters
        ----------
        inner : astropy.units.Quantity[angle]
            inner opening angle of cone
        outer : astropy.units.Quantity[angle]
            outer opening angle of cone

        Returns
        -------
        integrated : PowerLaw
            A new powerlaw instance with new normalization with the integration
            result.
        """
        if not self.normalization.unit.is_equivalent(DIFFUSE_FLUX_UNIT):
            raise ValueError("Can only integrate a diffuse flux over solid angle")

        solid_angle = cone_solid_angle(outer) - cone_solid_angle(inner)

        return PowerLaw(
            normalization=self.normalization * solid_angle,
            index=self.index,
            e_ref=self.e_ref,
        )



class LogParabola:
    r"""
    A log parabola flux parameterization.

    .. math::

        \Phi(E, \Phi_0, \alpha, \beta, E_\text{ref}) =
        \Phi_0 \left(
            \frac{E}{E_\text{ref}}
        \right)^{\alpha + \beta \cdot \log_{10}(E / E_\text{ref})}

    Attributes
    ----------
    normalization: astropy.units.Quantity[flux]
        :math:`\Phi_0`,
    a: float
        :math:`\alpha`
    b: float
        :math:`\beta`
    e_ref: astropy.units.Quantity[energy]
        :math:`E_\text{ref}`
    """

    @u.quantity_input(
        normalization=[DIFFUSE_FLUX_UNIT, POINT_SOURCE_FLUX_UNIT], e_ref=u.TeV
    )
    def __init__(self, normalization, a, b, e_ref=1 * u.TeV):
        """Create a new LogParabola spectrum"""
        self.normalization = normalization
        self.a = a
        self.b = b
        self.e_ref = e_ref

    @u.quantity_input(energy=u.TeV)
    def __call__(self, energy):
        e = (energy / self.e_ref).to_value(u.one)
        return self.normalization * e ** (self.a + self.b * np.log10(e))

    def __repr__(self):
        return f"{self.__class__.__name__}({self.normalization} * (E / {self.e_ref})**({self.a} + {self.b} * log10(E / {self.e_ref}))"


class PowerLawWithExponentialGaussian(PowerLaw):
    r"""
    A power law with an additional Gaussian bump.
    Beware that the Gaussian is not normalized!

    .. math::

        \Phi(E, \Phi_0, \gamma, f, \mu, \sigma, E_\text{ref}) =
        \Phi_0 \left(
            \frac{E}{E_\text{ref}}
        \right)^{\gamma}
        \cdot \left(
            1 + f \cdot
            \left(
                \exp\left(
                    \operatorname{Gauss}(\log_{10}(E / E_\text{ref}), \mu, \sigma)
                \right) - 1
            \right)
        \right)

    Where :math:`\operatorname{Gauss}` is the unnormalized Gaussian distribution:

    .. math::
        \operatorname{Gauss}(x, \mu, \sigma) = \exp\left(
            -\frac{1}{2} \left(\frac{x - \mu}{\sigma}\right)^2
        \right)

    Attributes
    ----------
    normalization: astropy.units.Quantity[flux]
        :math:`\Phi_0`,
    a: float
        :math:`\alpha`
    b: float
        :math:`\beta`
    e_ref: astropy.units.Quantity[energy]
        :math:`E_\text{ref}`
    """

    @u.quantity_input(
        normalization=[DIFFUSE_FLUX_UNIT, POINT_SOURCE_FLUX_UNIT], e_ref=u.TeV
    )
    def __init__(self, normalization, index, e_ref, f, mu, sigma):
        """Create a new PowerLawWithExponentialGaussian spectrum"""
        super().__init__(normalization=normalization, index=index, e_ref=e_ref)
        self.f = f
        self.mu = mu
        self.sigma = sigma

    @u.quantity_input(energy=u.TeV)
    def __call__(self, energy):
        power = super().__call__(energy)
        log10_e = np.log10(energy / self.e_ref)
        # ROOT's TMath::Gauss does not add the normalization
        # this is missing from the IRFDocs
        # the code used for the plot can be found here:
        # https://gitlab.cta-observatory.org/cta-consortium/aswg/irfs-macros/cosmic-rays-spectra/-/blob/master/electron_spectrum.C#L508
        gauss = np.exp(-0.5 * ((log10_e - self.mu) / self.sigma) ** 2)
        return power * (1 + self.f * (np.exp(gauss) - 1))

    def __repr__(self):
        s = super().__repr__()
        gauss = f"Gauss(log10(E / {self.e_ref}), {self.mu}, {self.sigma})"
        return s[:-1] + f" * (1 + {self.f} * (exp({gauss}) - 1))"


class TableInterpolationSpectrum:
    """
    Interpolate flux points to obtain a spectrum.

    By default, flux is interpolated linearly in log-log space.
    """

    def __init__(
        self, energy, flux, log_energy=True, log_flux=True, reference_energy=1 * u.TeV
    ):
        """Create a new TableInterpolationSpectrum spectrum"""
        self.energy = energy
        self.flux = flux
        self.flux_unit = flux.unit
        self.log_energy = log_energy
        self.log_flux = log_flux
        self.reference_energy = reference_energy

        x = (energy / reference_energy).to_value(u.one)
        y = flux.to_value(self.flux_unit)

        if log_energy:
            x = np.log10(x)

        if log_flux:
            y = np.log10(y)

        self.interp = interp1d(x, y, bounds_error=False, fill_value="extrapolate")

    def __call__(self, energy):

        x = (energy / self.reference_energy).to_value(u.one)

        if self.log_energy:
            x = np.log10(x)

        y = self.interp(x)

        if self.log_flux:
            y = 10 ** y

        return u.Quantity(y, self.flux_unit, copy=False)

    @classmethod
    def from_table(
        cls, table: QTable, log_energy=True, log_flux=True, reference_energy=1 * u.TeV
    ):
        return cls(
            table["energy"],
            table["flux"],
            log_energy=log_energy,
            log_flux=log_flux,
            reference_energy=reference_energy,
        )

    @classmethod
    def from_file(
        cls, path, log_energy=True, log_flux=True, reference_energy=1 * u.TeV
    ):
        return cls.from_table(
            QTable.read(path),
            log_energy=log_energy,
            log_flux=log_flux,
            reference_energy=reference_energy,
        )


#: Power Law parametrization of the Crab Nebula spectrum as published by HEGRA
#:
#: From "The Crab Nebula and Pulsar between 500 GeV and 80 TeV: Observations with the HEGRA stereoscopic air Cherenkov telescopes",
#: Aharonian et al, 2004, ApJ 614.2
#: doi.org/10.1086/423931
CRAB_HEGRA = PowerLaw(
    normalization=2.83e-11 / (u.TeV * u.cm ** 2 * u.s), index=-2.62, e_ref=1 * u.TeV,
)

#: Log-Parabola parametrization of the Crab Nebula spectrum as published by MAGIC
#:
#: From "Measurement of the Crab Nebula spectrum over three decades in energy with the MAGIC telescopes",
#: Aleksìc et al., 2015, JHEAP
#: https://doi.org/10.1016/j.jheap.2015.01.002
CRAB_MAGIC_JHEAP2015 = LogParabola(
    normalization=3.23e-11 / (u.TeV * u.cm ** 2 * u.s), a=-2.47, b=-0.24,
)


#: All particle spectrum
#:
#: (30.2) from "The Review of Particle Physics (2020)"
#: https://pdg.lbl.gov/2020/reviews/rpp2020-rev-cosmic-rays.pdf
PDG_ALL_PARTICLE = PowerLaw(
    normalization=1.8e4 / (u.GeV * u.m ** 2 * u.s * u.sr), index=-2.7, e_ref=1 * u.GeV,
)

#: Proton spectrum definition defined in the CTA Prod3b IRF Document
#:
#: From "Description of CTA Instrument Response Functions (Production 3b Simulation)", section 4.3.1
#: https://gitlab.cta-observatory.org/cta-consortium/aswg/documentation/internal_reports/irfs-reports/prod3b-irf-description
IRFDOC_PROTON_SPECTRUM = PowerLaw(
    normalization=9.8e-6 / (u.cm ** 2 * u.s * u.TeV * u.sr),
    index=-2.62,
    e_ref=1 * u.TeV,
)

#: Electron spectrum definition defined in the CTA Prod3b IRF Document
#:
#: From "Description of CTA Instrument Response Functions (Production 3b Simulation)", section 4.3.1
#: https://gitlab.cta-observatory.org/cta-consortium/aswg/documentation/internal_reports/irfs-reports/prod3b-irf-description
IRFDOC_ELECTRON_SPECTRUM = PowerLawWithExponentialGaussian(
    normalization=2.385e-9 / (u.TeV * u.cm ** 2 * u.s * u.sr),
    index=-3.43,
    e_ref=1 * u.TeV,
    mu=-0.101,
    sigma=0.741,
    f=1.950,
)

#: Proton + Helium interpolated from DAMPE measurements
#:
#: Datapoints obtained from obtained from:
#: https://inspirehep.net/files/62efc8374ffced58ea7e3a333bfa1217
#: Points are from DAMPE, up to  8 TeV.
#: For higher energies we assume a
#: flattening of the dF/dE*E^2.7 more or less in the middle of the large
#: spread of the available data reported on the same proceeding.
DAMPE_P_He_SPECTRUM = TableInterpolationSpectrum.from_file(
    resource_filename("pyirf", "resources/dampe_p+he.ecsv")
)



'''
Test export to GADF format
'''
import astropy.units as u
import numpy as np
from astropy.io import fits
import pytest
import tempfile


e_bins = np.geomspace(0.1, 100, 31) * u.TeV
migra_bins = np.linspace(0.2, 5, 101)
fov_bins = [0, 1, 2, 3] * u.deg
source_bins = np.linspace(0, 1, 101) * u.deg


@pytest.fixture
def aeff2d_hdus():
    from pyirf.io import create_aeff2d_hdu

    area = np.full((len(e_bins) - 1, len(fov_bins) - 1), 1e6) * u.m**2

    hdus = [
        create_aeff2d_hdu(area, e_bins, fov_bins, point_like=point_like)
        for point_like in [True, False]
    ]

    return area, hdus


@pytest.fixture
def edisp_hdus():
    from pyirf.io import create_energy_dispersion_hdu

    edisp = np.zeros((len(e_bins) - 1, len(migra_bins) - 1, len(fov_bins) - 1))
    edisp[:, 50, :] = 1.0

    hdus = [
        create_energy_dispersion_hdu(
            edisp, e_bins, migra_bins, fov_bins, point_like=point_like
        )
        for point_like in [True, False]
    ]

    return edisp, hdus


@pytest.fixture
def psf_hdu():
    from pyirf.io import create_psf_table_hdu
    from pyirf.utils import cone_solid_angle

    psf = np.zeros((len(e_bins) - 1, len(fov_bins) - 1, len(source_bins) - 1))
    psf[:, 0, :] = 1
    psf = psf / cone_solid_angle(source_bins[1])

    hdu = create_psf_table_hdu(
        psf, e_bins, source_bins, fov_bins, point_like=False
    )
    return psf, hdu


@pytest.fixture
def bg_hdu():
    from pyirf.io import create_background_2d_hdu

    background = np.column_stack([
        np.geomspace(1e9, 1e3, len(e_bins) - 1),
        np.geomspace(0.5e9, 0.5e3, len(e_bins) - 1),
        np.geomspace(1e8, 1e2, len(e_bins) - 1),
    ]) * u.Unit('TeV-1 s-1 sr-1')

    hdu = create_background_2d_hdu(background, e_bins, fov_bins)

    return background, hdu


@pytest.fixture
def rad_max_hdu():
    from pyirf.io import create_rad_max_hdu

    rad_max = np.full((len(e_bins) - 1, len(fov_bins) - 1), 0.1) * u.deg
    hdu = create_rad_max_hdu(rad_max, e_bins, fov_bins)

    return rad_max, hdu


def test_effective_area2d_gammapy(aeff2d_hdus):
    '''Test our effective area is readable by gammapy'''
    from gammapy.irf import EffectiveAreaTable2D

    area, hdus = aeff2d_hdus

    for hdu in hdus:
        with tempfile.NamedTemporaryFile(suffix='.fits') as f:
            fits.HDUList([fits.PrimaryHDU(), hdu]).writeto(f.name)
            # test reading with gammapy works
            aeff2d = EffectiveAreaTable2D.read(f.name)
            assert u.allclose(area, aeff2d.quantity, atol=1e-16 * u.m**2)


def test_effective_area2d_schema(aeff2d_hdus):
    '''Test our effective area is readable by gammapy'''
    from ogadf_schema.irfs import AEFF_2D

    _, hdus = aeff2d_hdus

    for hdu in hdus:
        AEFF_2D.validate_hdu(hdu)


def test_energy_dispersion_gammapy(edisp_hdus):
    '''Test our energy dispersion is readable by gammapy'''
    from gammapy.irf import EnergyDispersion2D

    edisp, hdus = edisp_hdus

    for hdu in hdus:
        with tempfile.NamedTemporaryFile(suffix='.fits') as f:
            fits.HDUList([fits.PrimaryHDU(), hdu]).writeto(f.name)

            # test reading with gammapy works
            edisp2d = EnergyDispersion2D.read(f.name, 'EDISP')
            assert u.allclose(edisp, edisp2d.quantity, atol=1e-16)


def test_energy_dispersion_schema(edisp_hdus):
    from ogadf_schema.irfs import EDISP_2D

    _, hdus = edisp_hdus

    for hdu in hdus:
        EDISP_2D.validate_hdu(hdu)


def test_psf_table_gammapy(psf_hdu):
    '''Test our psf is readable by gammapy'''
    from gammapy.irf import PSF3D

    psf, hdu = psf_hdu

    with tempfile.NamedTemporaryFile(suffix='.fits') as f:
        fits.HDUList([fits.PrimaryHDU(), hdu]).writeto(f.name)

        # test reading with gammapy works
        psf3d = PSF3D.read(f.name, 'PSF')
        assert u.allclose(psf, psf3d.quantity, atol=1e-16 / u.sr)


def test_psf_schema(psf_hdu):
    from ogadf_schema.irfs import PSF_TABLE

    _, hdu = psf_hdu
    PSF_TABLE.validate_hdu(hdu)


def test_background_2d_gammapy(bg_hdu):
    '''Test our background hdu is readable by gammapy'''
    from gammapy.irf import Background2D

    background, hdu = bg_hdu

    with tempfile.NamedTemporaryFile(suffix='.fits') as f:
        fits.HDUList([fits.PrimaryHDU(), hdu]).writeto(f.name)

        # test reading with gammapy works
        bg2d = Background2D.read(f.name, 'BACKGROUND')

        assert u.allclose(background, bg2d.quantity, atol=1e-16 * u.Unit('TeV-1 s-1 sr-1'))


def test_background_2d_schema(bg_hdu):
    from ogadf_schema.irfs import BKG_2D

    _, hdu = bg_hdu
    BKG_2D.validate_hdu(hdu)


def test_rad_max_schema(rad_max_hdu):
    from ogadf_schema.irfs import RAD_MAX

    _, hdu = rad_max_hdu
    RAD_MAX.validate_hdu(hdu)



"""
Functions to calculate sensitivity
"""
import numpy as np
from scipy.optimize import brentq
import logging

from astropy.table import QTable
import astropy.units as u

from .statistics import li_ma_significance
from .utils import check_histograms, cone_solid_angle
from .binning import create_histogram_table, bin_center


__all__ = ["relative_sensitivity", "calculate_sensitivity"]


log = logging.getLogger(__name__)


def _relative_sensitivity(
    n_on,
    n_off,
    alpha,
    min_significance=5,
    min_signal_events=10,
    min_excess_over_background=0.05,
    significance_function=li_ma_significance,
):
    if np.isnan(n_on) or np.isnan(n_off):
        return np.nan

    if n_on < 0 or n_off < 0:
        raise ValueError(f'n_on and n_off must be positive, got {n_on}, {n_off}')

    n_background = n_off * alpha
    n_signal = n_on - n_background

    if n_signal <= 0:
        return np.inf

    def equation(relative_flux):
        n_on = n_signal * relative_flux + n_background
        s = significance_function(n_on, n_off, alpha)
        return s - min_significance

    try:
        # brentq needs a lower and an upper bound
        # we will use the simple, analytically  solvable significance formula and scale it
        # with 10 to be sure it's above the Li and Ma solution
        # so rel * n_signal / sqrt(n_background) = target_significance
        if n_off > 1:
            relative_flux_naive = min_significance * np.sqrt(n_background) / n_signal
            upper_bound = 10 * relative_flux_naive
            lower_bound = 0.01 * relative_flux_naive
        else:
            upper_bound = 100
            lower_bound = 1e-4

        relative_flux = brentq(equation, lower_bound, upper_bound)

    except (RuntimeError, ValueError) as e:
        log.warn(
            "Could not calculate relative significance for"
            f" n_signal={n_signal:.1f}, n_off={n_off:.1f}, returning nan {e}"
        )
        return np.nan

    # less than min_sigma
    if relative_flux > 1:
        return np.nan

    # scale to achieved flux level
    n_signal = n_signal * relative_flux
    min_excess = min_excess_over_background * n_background
    min_signal = max(min_signal_events, min_excess)

    # if we violate the min signal events condition,
    # increase flux until we meet the requirement
    if n_signal < min_signal:
        scale = min_signal / n_signal
    else:
        scale = 1.0

    return relative_flux * scale


_relative_sensitivity_vectorized = np.vectorize(
    _relative_sensitivity,
    excluded=['significance_function']
)


def relative_sensitivity(
    n_on,
    n_off,
    alpha,
    min_significance=5,
    min_signal_events=10,
    min_excess_over_background=0.05,
    significance_function=li_ma_significance,
):
    """
    Calculate the relative sensitivity defined as the flux
    relative to the reference source that is detectable with
    significance ``target_significance``.

    Given measured ``n_on`` and ``n_off``,
    we estimate the number of gamma events ``n_signal`` as ``n_on - alpha * n_off``.

    The number of background events ``n_background` is estimated as ``n_off * alpha``.

    In the end, we find the relative sensitivity as the scaling factor for ``n_signal``
    that yields a significance of ``target_significance``.

    The reference time should be incorporated by appropriately weighting the events
    before calculating ``n_on`` and ``n_off``.

    All input values with the exception of ``significance_function``
    must be broadcastable to a single, common shape.

    Parameters
    ----------
    n_on: int or array-like
        Number of signal-like events for the on observations
    n_off: int or array-like
        Number of signal-like events for the off observations
    alpha: float or array-like
        Scaling factor between on and off observations.
        1 / number of off regions for wobble observations.
    min_significance: float or array-like
        Significance necessary for a detection
    min_signal_events: int or array-like
        Minimum number of signal events required.
        The relative flux will be scaled up from the one yielding ``min_significance``
        if this condition is violated.
    min_excess_over_background: float or array-like
        Minimum number of signal events expressed as the proportion of the
        background.
        So the required number of signal events will be
        ``min_excess_over_background * alpha * n_off``.
        The relative flux will be scaled up from the one yielding ``min_significance``
        if this condition is violated.
    significance_function: function
        A function f(n_on, n_off, alpha) -> significance in sigma
        Used to calculate the significance, default is the Li&Ma
        likelihood ratio test formula.
        Li, T-P., and Y-Q. Ma.
        "Analysis methods for results in gamma-ray astronomy."
        The Astrophysical Journal 272 (1983): 317-324.
        Formula (17)
    """
    return _relative_sensitivity_vectorized(
        n_on=n_on,
        n_off=n_off,
        alpha=alpha,
        min_significance=min_significance,
        min_signal_events=min_signal_events,
        min_excess_over_background=min_excess_over_background,
        significance_function=significance_function,
    )


def calculate_sensitivity(
    signal_hist,
    background_hist,
    alpha,
    min_significance=5,
    min_signal_events=10,
    min_excess_over_background=0.05,
    significance_function=li_ma_significance,
):
    """
    Calculate sensitivity for DL2 event lists in bins of reconstructed energy.

    Sensitivity is defined as the minimum flux detectable with ``target_significance``
    sigma significance in a certain time.

    This time must be incorporated into the event weights.

    Two conditions are required for the sensitivity:
    - At least ten weighted signal events
    - The weighted signal must be larger than 5 % of the weighted background
    - At least 5 sigma (so relative_sensitivity > 1)

    If the conditions are not met, the sensitivity will be set to nan.

    Parameters
    ----------
    signal_hist: astropy.table.QTable
        Histogram of detected signal events as a table.
        Required columns: n and n_weighted.
        See ``pyirf.binning.create_histogram_table``
    background_hist: astropy.table.QTable
        Histogram of detected events as a table.
        Required columns: n and n_weighted.
        See ``pyirf.binning.create_histogram_table``
    alpha: float
        Size ratio of signal region to background region
    min_significance: float
        Significance necessary for a detection
    min_signal_events: int
        Minimum number of signal events required.
        The relative flux will be scaled up from the one yielding ``min_significance``
        if this condition is violated.
    min_excess_over_background: float
        Minimum number of signal events expressed as the proportion of the
        background.
        So the required number of signal events will be
        ``min_excess_over_background * alpha * n_off``.
        The relative flux will be scaled up from the one yielding ``min_significance``
        if this condition is violated.
    significance_function: callable
        A function with signature (n_on, n_off, alpha) -> significance.
        Default is the Li & Ma likelihood ratio test.

    Returns
    -------
    sensitivity_table: astropy.table.QTable
        Table with sensitivity information.
        Contains weighted and unweighted number of signal and background events
        and the ``relative_sensitivity``, the scaling applied to the signal events
        that yields ``target_significance`` sigma of significance according to
        the ``significance_function``
    """
    check_histograms(signal_hist, background_hist)

    n_on = signal_hist["n_weighted"] + alpha * background_hist["n_weighted"]

    # convert any quantities to arrays,
    # since quantitites don't work with vectorize
    n_on = u.Quantity(n_on, copy=False).to_value(u.one)
    n_off = u.Quantity(background_hist["n_weighted"], copy=False).to_value(u.one)

    # calculate sensitivity in each bin
    rel_sens = relative_sensitivity(
        n_on=n_on,
        n_off=n_off,
        alpha=alpha,
        min_significance=min_significance,
        min_signal_events=min_signal_events,
        min_excess_over_background=min_excess_over_background,
        significance_function=significance_function,
    )

    # fill output table
    s = QTable()
    for key in ("low", "high", "center"):
        k = "reco_energy_" + key
        s[k] = signal_hist[k]

    s["n_signal"] = signal_hist["n"] * rel_sens
    s["n_signal_weighted"] = signal_hist["n_weighted"] * rel_sens
    s["n_background"] = background_hist["n"]
    s["n_background_weighted"] = background_hist["n_weighted"]

    # copy also "n_proton" / "n_electron_weighted" etc. if available
    for k in filter(lambda c: c.startswith('n_') and c != 'n_weighted', background_hist.colnames):
        s[k] = background_hist[k]

    s["significance"] = significance_function(
        n_on=s["n_signal_weighted"] + alpha * s["n_background_weighted"],
        n_off=s["n_background_weighted"],
        alpha=alpha,
    )
    s["relative_sensitivity"] = rel_sens

    return s


def estimate_background(
    events, reco_energy_bins, theta_cuts, alpha,
    fov_offset_min, fov_offset_max, 
):
    '''
    Estimate the number of background events for a point-like sensitivity.

    Due to limited statistics, it is often not possible to just apply the same
    theta cut to the background events as to the signal events around an assumed
    source position.

    Here we calculate the expected number of background events for the off
    regions by taking all background events between ``fov_offset_min`` and
    ``fov_offset_max`` from the camera center and then scale these to the size
    of the off region, which is scaled by 1 / alpha from the size of the on
    region given by the theta cuts.


    Parameters
    ----------
    events: astropy.table.QTable
        DL2 event list of background surviving event selection
        and inside ``fov_offset_max`` from the center of the FOV
        Required columns for this function:
        - `reco_energy`,
        - `reco_source_fov_offset`.
    reco_energy_bins: astropy.units.Quantity[energy]
        Desired bin edges in reconstructed energy for the background rate
    theta_cuts: astropy.table.QTable
        The cuts table for the theta cut,
        e.g. as returned by ``~pyirf.cuts.calculate_percentile_cut``.
        Columns `center` and `cut` are required for this function.
    alpha: float
        size of the on region divided by the size of the off region.
    fov_offset_min: astropy.units.Quantity[angle]
        Minimum distance from the fov center for background events to be taken into account
    fov_offset_max: astropy.units.Quantity[angle]
        Maximum distance from the fov center for background events to be taken into account
    '''
    in_ring = (
        (events['reco_source_fov_offset'] >= fov_offset_min)
        & (events['reco_source_fov_offset'] < fov_offset_max)
    )

    bg = create_histogram_table(
        events[in_ring],
        reco_energy_bins,
        key='reco_energy',
    )

    # scale number of background events according to the on region size
    # background radius and alpha
    center = bin_center(reco_energy_bins)
    # interpolate the theta cut to the bins used here
    theta_cuts_bg_bins = np.interp(
        center,
        theta_cuts['center'],
        theta_cuts['cut']
    )

    solid_angle_on = cone_solid_angle(theta_cuts_bg_bins) 
    solid_angle_ring = cone_solid_angle(fov_offset_max) - cone_solid_angle(fov_offset_min)
    size_ratio = (solid_angle_on / solid_angle_ring).to_value(u.one)

    for key in filter(lambda col: col.startswith('n'), bg.colnames):
        # *= not possible due to upcast from int to float
        bg[key] = bg[key] * size_ratio / alpha

    return bg



"""Functions for performing interpolation of IRF to the values read from the data."""

import astropy.units as u
import numpy as np

from .griddata_interpolator import GridDataInterpolator
from .quantile_interpolator import QuantileInterpolator
from pyirf.utils import cone_solid_angle

__all__ = [
    "interpolate_effective_area_per_energy_and_fov",
    "interpolate_energy_dispersion",
    "interpolate_psf_table",
    "interpolate_rad_max",
]


@u.quantity_input(effective_area=u.m**2)
def interpolate_effective_area_per_energy_and_fov(
    effective_area,
    grid_points,
    target_point,
    min_effective_area=1 * u.m**2,
    method="linear",
):
    """
    Takes a grid of effective areas for a bunch of different parameters
    and interpolates (log) effective areas to given value of those parameters

    Parameters
    ----------
    effective_area: np.array of astropy.units.Quantity[area]
        grid of effective area, of shape (n_grid_points, n_fov_offset_bins, n_energy_bins)
    grid_points: np.array
        list of parameters corresponding to effective_area, of shape (n_grid_points, n_interp_dim)
    target_point: np.array
        values of parameters for which the interpolation is performed, of shape (n_interp_dim)
    min_effective_area: astropy.units.Quantity[area]
        Minimum value of effective area to be considered for interpolation
    method: 'linear’, ‘nearest’, ‘cubic’
        Interpolation method

    Returns
    -------
    aeff_interp: astropy.units.Quantity[area]
        Interpolated Effective area array with shape (n_energy_bins, n_fov_offset_bins)
    """

    # get rid of units
    effective_area = effective_area.to_value(u.m**2)
    min_effective_area = min_effective_area.to_value(u.m**2)

    # remove zeros and log it
    effective_area[effective_area < min_effective_area] = min_effective_area
    effective_area = np.log(effective_area)

    # interpolation
    interp = GridDataInterpolator(
        grid_points=grid_points,
        params=effective_area,
    )
    aeff_interp = interp(target_point, method=method)
    # exp it and set to zero too low values
    aeff_interp = np.exp(aeff_interp)
    # 1.1 to correct for numerical uncertainty and interpolation
    aeff_interp[aeff_interp < min_effective_area * 1.1] = 0
    return u.Quantity(aeff_interp, u.m**2, copy=False)


def interpolate_energy_dispersion(
    migra_bins, edisps, grid_points, target_point, quantile_resolution=1e-3
):
    """
    Takes a grid of energy dispersions for a bunch of different parameters
    and interpolates it to given value of those parameters.

    Parameters
    ----------
    migra_bins: numpy.ndarray, shape=(M+1)
        Common array of migration bin-edges

    edisps: numpy.ndarray, shape=(N, ..., M,...)
        The EDISP MATRIX, shape is assumed to be (N:n_grid_points, n_energy_bins, M:n_migration_bins, n_fov_offset_bins).

    grid_points: numpy.ndarray, shape=(N, O)
        Array of the N O-dimensional morphing parameter values corresponding to the N input templates. The pdf's quantiles
        are expected to vary linearly between these two reference points.

    target_point: numpy.ndarray, shape=(O)
        Value for which the interpolation is performed (target point)

    quantile_resolution: float
        Spacing between the quantiles that are computed in the interpolation. Defaults to 1e-3.

    Returns
    -------
    edisp_interp: numpy.ndarray, shape=(1,...,M,...)
        Interpolated and binned energy dispersion.
    """
    interp = QuantileInterpolator(
        bin_edges=migra_bins,
        bin_contents=edisps,
        grid_points=grid_points,
        axis=-2,
        quantile_resolution=quantile_resolution,
    )

    return interp(target_point)


@u.quantity_input(psf=u.sr**-1, source_offset_bins=u.deg)
def interpolate_psf_table(
    source_offset_bins, psfs, grid_points, target_point, quantile_resolution=1e-3
):
    """
    Takes a grid of PSF tables for a bunch of different parameters
    and interpolates it to given value of those parameters

    Parameters
    ----------
    source_offset_bins: numpy.ndarray, shape=(M+1), of astropy.units.Quantity[deg]
        Common array of source offset bin-edges

    psfs: numpy.ndarray, shape=(N, ..., M), of astropy.units.Quantity[sr**-1]
        The PSF_TABLE, shape is assumed to be (N:n_grid_points, n_energy_bins, n_fov_offset_bins, M:n_source_offset_bins).

    grid_points: numpy.ndarray, shape=(N, O)
        Array of the N O-dimensional morphing parameter values corresponding to the N input templates. The pdf's quantiles
        are expected to vary linearly between these two reference points.

    target_point: numpy.ndarray, shape=(O)
        Value for which the interpolation is performed (target point)

    quantile_resolution: float
        Spacing between the quantiles that are computed in the interpolation. Defaults to 1e-3.

    Returns
    -------
    psf_interp: numpy.ndarray, shape=(1, ..., M)
        Interpolated PSF table with shape (n_energy_bins, n_fov_offset_bins, n_source_offset_bins)
    """

    # Renormalize along the source offset axis to have a proper PDF
    omegas = np.diff(cone_solid_angle(source_offset_bins))
    psfs_normed = psfs * omegas

    interp = QuantileInterpolator(
        bin_edges=source_offset_bins,
        bin_contents=psfs_normed,
        grid_points=grid_points,
        axis=-1,
        quantile_resolution=quantile_resolution,
    )

    interpolated_psf_normed = interp(target_point)

    # Undo normalisation to get a proper PSF and return
    return interpolated_psf_normed / omegas


def interpolate_rad_max(
    rad_max,
    grid_points,
    target_point,
    method="linear",
):
    """
    Interpolates a grid of RAD_MAX tables for point-like IRFs to a target-point.
    Wrapper around scipy.interpolate.griddata [1].

    Parameters
    ----------
    rad_max: numpy.ndarray, shape=(N, M, ...)
        Theta-cuts for all combinations of grid-points, energy and fov_offset.
        Shape (N:n_grid_points, M:n_energy_bins, n_fov_offset_bins)

    grid_points: numpy.ndarray, shape=(N, O)
        Array of the N O-dimensional morphing parameter values corresponding to the N input templates.

    target_point: numpy.ndarray, shape=(O)
        Value for which the interpolation is performed (target point)

    method: 'linear', 'nearest', 'cubic'
        Interpolation method for scipy.interpolate.griddata [1]. Defaults to 'linear'.

    Returns
    -------
    rad_max_interp: numpy.ndarray, shape=(1, M, ...)
        Theta-cuts for the target grid-point, shape (1, M:n_energy_bins, n_fov_offset_bins)

    References
    ----------
    .. [1] https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.griddata.html
    """

    interp = GridDataInterpolator(grid_points=grid_points, params=rad_max)
    return interp(target_point, method=method)



from setuptools import setup, find_packages
import os


gammapy = "gammapy~=1.0"

extras_require = {
    "docs": [
        "sphinx",
        "sphinx_rtd_theme",
        "sphinx_automodapi",
        "numpydoc",
        "nbsphinx",
        "uproot",
        "awkward1",
        "notebook",
        "tables",
        "towncrier",
        gammapy,
    ],
    "tests": [
        "pytest",
        "pytest-cov",
        gammapy,
        "ogadf-schema~=0.2.3",
        "uproot~=4.0",
        "awkward~=1.0",
    ],
    "gammapy": [
        gammapy,
    ],
}
extras_require["dev"] = extras_require["tests"] + [
    "setuptools_scm",
]

all_extras = set()
for extra in extras_require.values():
    all_extras.update(extra)
extras_require["all"] = list(all_extras)

setup(
    use_scm_version={"write_to": os.path.join("pyirf", "_version.py")},
    packages=find_packages(exclude=['pyirf._dev_version']),
    install_requires=[
        "astropy>=4.0.2",
        "matplotlib",
        "numpy>=1.18",
        "scipy",
        "tqdm",
    ],
    include_package_data=True,
    extras_require=extras_require,
)



import numpy as np
import pytest
from scipy.stats import norm

from pyirf.binning import bin_center


@pytest.fixture
def data():
    """Create common dataset containing binned Gaussians for interpolation testing. Binned PDFs sum to 1."""
    bin_edges = np.linspace(-5, 30, 101)
    distributions = [norm(5, 1), norm(10, 2), norm(15, 3)]

    # create binned pdfs by interpolation of bin content
    binned_pdfs = np.array([np.diff(dist.cdf(bin_edges)) for dist in distributions])

    dataset = {
        "bin_edges": bin_edges,
        "means": np.array([5, 10, 15]),
        "stds": np.array([1, 2, 3]),
        "distributions": distributions,
        "binned_pdfs": binned_pdfs,
        "grid_points": np.array([1, 2, 3]),
    }

    return dataset


def test_cdf_values(data):
    from pyirf.interpolation.quantile_interpolator import cdf_values

    cdf_est = cdf_values(data["binned_pdfs"][0])

    # Assert empty histograms result in cdf containing zeros
    assert np.all(cdf_values(np.zeros(shape=5)) == 0)

    # Assert cdf is increasing or constant for actual pdfs
    assert np.all(np.diff(cdf_est) >= 0)

    # Assert cdf is capped at 1
    assert np.max(cdf_est) == 1

    # Assert estimated and true cdf are matching
    assert np.allclose(cdf_est, data["distributions"][0].cdf(data["bin_edges"][1:]))


def test_ppf_values(data):
    from pyirf.interpolation.quantile_interpolator import (cdf_values,
                                                           ppf_values)

    # Create quantiles, ignore the 0% and 100% quantile as they are analytically +- inf
    quantiles = np.linspace(0, 1, 10)[1:-2]

    # True ppf-values
    ppf_true = data["distributions"][0].ppf(quantiles)
    bin_mids = bin_center(data["bin_edges"])

    # Estimated ppf-values
    cdf_est = cdf_values(data["binned_pdfs"][0])
    ppf_est = ppf_values(bin_mids, cdf_est, quantiles)

    # Assert truth and estimation match allowing for +- bin_width deviation
    assert np.allclose(ppf_true, ppf_est, atol=np.diff(data["bin_edges"])[0])


def test_pdf_from_ppf(data):
    from pyirf.interpolation.quantile_interpolator import (cdf_values,
                                                           pdf_from_ppf,
                                                           ppf_values)

    # Create quantiles
    quantiles = np.linspace(0, 1, 1000)
    bin_mids = bin_center(data["bin_edges"])

    # Estimate ppf-values
    cdf_est = cdf_values(data["binned_pdfs"][0])
    ppf_est = ppf_values(bin_mids, cdf_est, quantiles)

    # Compute pdf_values
    pdf_est = pdf_from_ppf(data["bin_edges"], ppf_est, quantiles)

    # Assert pdf-values matching true pdf within +-1%
    assert np.allclose(pdf_est, data["binned_pdfs"][0], atol=1e-2)


def test_norm_pdf(data):
    from pyirf.interpolation.quantile_interpolator import norm_pdf

    assert np.allclose(norm_pdf(2 * data["binned_pdfs"][0]), data["binned_pdfs"][0])
    assert np.allclose(norm_pdf(np.zeros(5)), 0)


def test_interpolate_binned_pdf(data):
    from pyirf.interpolation import QuantileInterpolator

    interpolator = QuantileInterpolator(
        grid_points=data["grid_points"][[0, 2]],
        bin_edges=data["bin_edges"],
        bin_contents=data["binned_pdfs"][[0, 2], :],
        quantile_resolution=1e-3,
        axis=-1,
    )

    interp = interpolator(
        target_point=np.array([data["grid_points"][1]]),
    ).squeeze()

    bin_mids = bin_center(data["bin_edges"])
    bin_width = np.diff(data["bin_edges"])[0]

    # Estimate mean and standart_deviation from interpolant
    interp_mean = np.average(bin_mids, weights=interp)
    interp_std = np.sqrt(np.average((bin_mids - interp_mean) ** 2, weights=interp))

    # Assert they match the truth within one bin of uncertainty
    assert np.isclose(interp_mean, data["means"][1], atol=bin_width)
    assert np.isclose(interp_std, data["stds"][1], atol=bin_width)



import pytest
import astropy.units as u
import numpy as np

pytest.importorskip('gammapy')


true_energy_bins = [0.1, 1, 10, 100] * u.TeV
fov_offset_bins = [0, 1, 2] * u.deg
source_offset_bins = np.linspace(0, 1, 20) * u.deg
migration_bins = np.geomspace(0.2, 5, 10)


def test_effective_area_table_2d():
    from pyirf.gammapy import create_effective_area_table_2d

    shape = (len(true_energy_bins) - 1, len(fov_offset_bins) - 1)
    aeff = np.random.uniform(0, 1e5, size=shape) * u.m**2

    aeff_gammapy = create_effective_area_table_2d(aeff, true_energy_bins, fov_offset_bins)
    # test str repr works

    str(aeff_gammapy)


def test_psf_3d():
    from pyirf.gammapy import create_psf_3d

    shape = (len(true_energy_bins) - 1, len(fov_offset_bins) - 1, len(source_offset_bins) - 1)
    psf = np.zeros(shape) / u.sr
    psf3d = create_psf_3d(psf, true_energy_bins, source_offset_bins,  fov_offset_bins)
    str(psf3d)


def test_energy_dispersion():
    from pyirf.gammapy import create_energy_dispersion_2d

    shape = (len(true_energy_bins) - 1, len(migration_bins) - 1, len(fov_offset_bins) - 1)
    edisp = np.zeros(shape)
    edisp2d = create_energy_dispersion_2d(edisp, true_energy_bins, migration_bins, fov_offset_bins)
    str(edisp2d)



import operator
import numpy as np
from astropy.table import QTable
import astropy.units as u
import pytest
from scipy.stats import norm


@pytest.fixture
def events():
    return QTable(
        {
            "bin_reco_energy": [0, 0, 1, 1, 2, 2],
            "theta": [0.1, 0.02, 0.3, 0.15, 0.01, 0.1] * u.deg,
            "gh_score": [1.0, -0.2, 0.5, 0.05, 1.0, 0.3],
        }
    )


def test_calculate_percentile_cuts():
    from pyirf.cuts import calculate_percentile_cut

    np.random.seed(0)

    dist1 = norm(0, 1)
    dist2 = norm(10, 1)
    N = int(1e4)

    values = np.append(dist1.rvs(size=N), dist2.rvs(size=N)) * u.deg
    bin_values = np.append(np.zeros(N), np.ones(N)) * u.m
    # add some values outside of binning to test that under/overflow are ignored
    bin_values[10] = 5 * u.m
    bin_values[30] = -1 * u.m

    bins = [-0.5, 0.5, 1.5] * u.m

    cuts = calculate_percentile_cut(values, bin_values, bins, fill_value=np.nan * u.deg)
    assert np.all(cuts["low"] == bins[:-1])
    assert np.all(cuts["high"] == bins[1:])

    assert np.allclose(
        cuts["cut"].to_value(u.deg),
        [dist1.ppf(0.68), dist2.ppf(0.68)],
        rtol=0.1,
    )

    # test with min/max value
    cuts = calculate_percentile_cut(
        values,
        bin_values,
        bins,
        fill_value=np.nan * u.deg,
        min_value=1 * u.deg,
        max_value=5 * u.deg,
    )
    assert np.all(cuts["cut"] == [1.0, 5.0] * u.deg)


def test_calculate_percentile_cuts_no_units():
    from pyirf.cuts import calculate_percentile_cut

    np.random.seed(0)

    dist1 = norm(0, 1)
    dist2 = norm(10, 1)
    N = int(1e4)

    values = np.append(dist1.rvs(size=N), dist2.rvs(size=N))
    bin_values = np.append(np.zeros(N), np.ones(N)) * u.m
    bins = [-0.5, 0.5, 1.5] * u.m

    cuts = calculate_percentile_cut(values, bin_values, bins, fill_value=np.nan)
    assert np.all(cuts["low"] == bins[:-1])
    assert np.all(cuts["high"] == bins[1:])

    assert np.allclose(
        cuts["cut"],
        [dist1.ppf(0.68), dist2.ppf(0.68)],
        rtol=0.1,
    )


def test_calculate_percentile_cuts_smoothing():
    from pyirf.cuts import calculate_percentile_cut

    np.random.seed(0)

    dist1 = norm(0, 1)
    dist2 = norm(10, 1)
    N = int(1e4)

    values = np.append(dist1.rvs(size=N), dist2.rvs(size=N))
    bin_values = np.append(np.zeros(N), np.ones(N)) * u.m
    bins = [-0.5, 0.5, 1.5] * u.m

    cuts = calculate_percentile_cut(values, bin_values, bins, fill_value=np.nan, smoothing=1)
    assert np.all(cuts["low"] == bins[:-1])
    assert np.all(cuts["high"] == bins[1:])

    assert np.allclose(
        cuts["cut"],
        [3.5, 7.5],
        rtol=0.2,
    )


def test_evaluate_binned_cut():
    from pyirf.cuts import evaluate_binned_cut

    cuts = QTable({"low": [0, 1], "high": [1, 2], "cut": [100, 1000],})

    survived = evaluate_binned_cut(
        np.array([500, 1500, 50, 2000, 25, 800]),
        np.array([0.5, 1.5, 0.5, 1.5, 0.5, 1.5]),
        cut_table=cuts,
        op=operator.ge,
    )
    assert np.all(survived == [True, True, False, True, False, False])

    # test with quantity
    cuts = QTable(
        {"low": [0, 1] * u.TeV, "high": [1, 2] * u.TeV, "cut": [100, 1000] * u.m,}
    )

    survived = evaluate_binned_cut(
        [500, 1500, 50, 2000, 25, 800] * u.m,
        [0.5, 1.5, 0.5, 1.5, 0.5, 1.5] * u.TeV,
        cut_table=cuts,
        op=operator.ge,
    )
    assert np.all(survived == [True, True, False, True, False, False])


def test_compare_irf_cuts():
    """Tests compare_irf_cuts."""

    from pyirf.cuts import compare_irf_cuts
    # first create some dummy cuts
    enbins = np.logspace(-2, 3) * u.TeV
    thcuts1 = np.linspace(0.5, 0.1) * u.deg
    thcuts2 = np.linspace(0.6, 0.2) * u.deg
    names = ("Energy", "Theta2")
    t1 = QTable([enbins, thcuts1], names=names)
    t1b = QTable([enbins, thcuts1], names=names)
    t2 = QTable([enbins, thcuts2], names=names)
    # comparing identical cuts ==> should return True
    assert compare_irf_cuts([t1, t1b])

    # different cuts ==> should return False
    assert compare_irf_cuts([t1, t2]) is False


def test_calculate_percentile_cuts_table():
    '''Test that calculate percentile cuts does not modify input table'''
    from pyirf.cuts import calculate_percentile_cut

    np.random.seed(0)

    dist1 = norm(0, 1)
    dist2 = norm(10, 1)
    N = int(1e4)

    table = QTable({
        "foo": np.append(dist1.rvs(size=N), dist2.rvs(size=N)) * u.deg,
        "bar": np.append(np.zeros(N), np.ones(N)) * u.m,
    })

    bins = [-0.5, 0.5, 1.5] * u.m
    cuts = calculate_percentile_cut(
        table["foo"],
        table["bar"],
        bins,
        fill_value=np.nan * u.rad,
    )
    assert table.colnames == ["foo", "bar"]
    assert np.allclose(
        cuts["cut"].to_value(u.deg),
        [dist1.ppf(0.68), dist2.ppf(0.68)],
        rtol=0.1,
    )



import astropy.units as u
import numpy as np

from ..utils import cone_solid_angle

#: Unit of the background rate IRF
BACKGROUND_UNIT = u.Unit('s-1 TeV-1 sr-1')


def background_2d(events, reco_energy_bins, fov_offset_bins, t_obs):
    """
    Calculate background rates in radially symmetric bins in the field of view.

    GADF documentation here:
    https://gamma-astro-data-formats.readthedocs.io/en/latest/irfs/full_enclosure/bkg/index.html#bkg-2d

    Parameters
    ----------
    events: astropy.table.QTable
        DL2 events table of the selected background events.
        Needed columns for this function: `reco_source_fov_offset`, `reco_energy`, `weight`
    reco_energy: astropy.units.Quantity[energy]
        The bins in reconstructed energy to be used for the IRF
    fov_offset_bins: astropy.units.Quantity[angle]
        The bins in the field of view offset to be used for the IRF
    t_obs: astropy.units.Quantity[time]
        Observation time. This must match with how the individual event
        weights are calculated.

    Returns
    -------
    bg_rate: astropy.units.Quantity
        The background rate as particles per energy, time and solid angle
        in the specified bins.

        Shape: (len(reco_energy_bins) - 1, len(fov_offset_bins) - 1)
    """

    hist, _, _ = np.histogram2d(
        events["reco_energy"].to_value(u.TeV),
        events["reco_source_fov_offset"].to_value(u.deg),
        bins=[
            reco_energy_bins.to_value(u.TeV),
            fov_offset_bins.to_value(u.deg),
        ],
        weights=events['weight'],
    )

    # divide all energy bins by their width
    # hist has shape (n_energy, n_fov_offset) so we need to transpose and then back
    bin_width_energy = np.diff(reco_energy_bins)
    per_energy = (hist.T / bin_width_energy).T

    # divide by solid angle in each fov bin and the observation time
    bin_solid_angle = np.diff(cone_solid_angle(fov_offset_bins))
    bg_rate = per_energy / t_obs / bin_solid_angle

    return bg_rate.to(BACKGROUND_UNIT)



from astropy.table import QTable
import numpy as np
import astropy.units as u


def test_relative_sensitivity():
    from pyirf.sensitivity import relative_sensitivity

    # some general case
    n_on = 100
    n_off = 200
    alpha = 0.2
    assert 0.1 < relative_sensitivity(n_on, n_off, alpha) < 1.0

    # numbers yield lima = 5 relatively precisely, so sensitivity should be 1
    assert np.isclose(relative_sensitivity(81, 202, 0.2), 1, rtol=0.01)

    # test different target significance
    # numbers yield lima = 8 relatively precisely, so sensitivity should be 1
    result = relative_sensitivity(93, 150, 0.2, min_significance=8)
    assert np.isclose(result, 1, rtol=0.01)

    # no signal => inf
    assert np.isinf(relative_sensitivity(10, 100, 0.2))

    # no background, should work
    assert relative_sensitivity(10, 0, 0.2) > 0

    # test vectorization

    n_on = np.array([10, 20, 30])
    n_off = np.array([10, 100])
    s = relative_sensitivity(n_on, n_off[:, np.newaxis], alpha=0.2)
    assert s.shape == (2, 3)


def test_calculate_sensitivity():
    from pyirf.sensitivity import calculate_sensitivity
    from pyirf.binning import bin_center
    bins = [0.1, 1.0, 10] * u.TeV
    signal_hist = QTable({
        'reco_energy_low': bins[:-1],
        'reco_energy_high': bins[1:],
        'reco_energy_center': bin_center(bins),
        'n': [100, 100],
        'n_weighted': [100, 100],
    })
    bg_hist = signal_hist.copy()
    bg_hist['n'] = 20
    bg_hist['n_weighted'] = 50

    sensitivity = calculate_sensitivity(signal_hist, bg_hist, alpha=0.2)

    # too small sensitivity
    signal_hist['n_weighted'] = [10, 10]
    sensitivity = calculate_sensitivity(signal_hist, bg_hist, alpha=0.2)

    assert len(sensitivity) == len(signal_hist)
    assert np.all(np.isnan(sensitivity['relative_sensitivity']))

    # not above 5 percent of remaining background
    signal_hist['n_weighted'] = 699
    bg_hist['n_weighted'] = 70_000
    sensitivity = calculate_sensitivity(signal_hist, bg_hist, alpha=0.2)
    assert len(sensitivity) == len(signal_hist)
    # we scale up the signal until we meet the requirement, so
    # we must have more than 5 sigma
    assert np.all(sensitivity['significance'] >= 5)

    # less then 10 events
    signal_hist['n_weighted'] = [9, 9]
    bg_hist['n_weighted'] = [1, 1]
    sensitivity = calculate_sensitivity(signal_hist, bg_hist, alpha=0.2)
    # we scale up the signal until we meet the requirement, so
    # we must have more than 5 sigma
    assert np.all(sensitivity['significance'] >= 5)


def test_estimate_background():
    from pyirf.sensitivity import estimate_background
    N = 1000
    events = QTable({
        'reco_source_fov_offset': np.append(np.full(N, 0.5), np.full(N, 1.5)) * u.deg,
        'reco_energy': np.tile([5, 50], N) * u.TeV,
        'weight': np.tile([1, 2], N),
    })
    reco_energy_bins = [1, 10, 100] * u.TeV
    theta_cuts = QTable({
        'low': [1, 10] * u.TeV,
        'high': [10, 100] * u.TeV,
        'center': [5.5, 55] * u.TeV,
        'cut': (np.arccos([0.9998, 0.9999]) * u.rad).to(u.deg),
    })
    fov_offset_max = np.arccos(0.999) * u.rad

    bg = estimate_background(
        events,
        reco_energy_bins,
        theta_cuts,
        alpha=0.2,
        fov_offset_min=0 * u.deg,
        fov_offset_max=fov_offset_max,
    )

    assert np.allclose(bg['n'], [1000, 500])
    assert np.allclose(bg['n_weighted'], [1000, 1000])
